import datetime as dt
import re
import typing
import urllib
import urllib.parse
from dataclasses import dataclass, fields
from enum import Enum


class LogLineIndex(Enum):
    Raw = 0
    IpAddress = 1
    Identd = 2
    UserId = 3
    ReceivedAt = 4
    Url = 5
    StatusCode = 6
    ReturnSize = 7
    Referer = 8
    Agent = 9


@dataclass(frozen=True)
@dataclass
class UrlMeta:
    method: str
    path: str
    params: dict
    potocol: str
    metaStr: str

    @staticmethod
    def extractUrlMetaSlient(metaStr: str) -> typing.Union['UrlMeta', None]:
        try:
            return UrlMeta.extractUrlMeta(metaStr)
        except:
            return None

    @staticmethod
    def extractUrlMeta(metaStr: str) -> 'UrlMeta':
        method = ""
        path = ""
        param = dict()
        potocol = ""

        line = metaStr if type(metaStr) is not tuple else metaStr[0]

        if line not in ['-', "\\n"]:
            # v1 => method
            # v2 => path
            # v3 => potocol
            v1, v2, v3 = line.split(" ")
            method = urllib.parse.unquote(v1)
            vPath = v2.split("?")

            url = None
            if re.match(urllib.parse.unquote_plus(vPath[0]), "[a-zA-Z0-9]{4}"):
                url = urllib.parse.unquote_plus(url)
            else:
                url = urllib.parse.unquote_plus(vPath[0])

            path = url

            v5d = {
                (urllib.parse.unquote(pair[0]), urllib.parse.unquote(pair[1])) if len(pair) == 2 else (
                    urllib.parse.unquote(pair[0]), "")
                for paramStr in vPath[1].split("&") if len(paramStr) > 0
                for pair in [paramStr.split("=")]

            } if len(vPath) == 2 else {}

            param = v5d
            potocol = urllib.parse.unquote(v3)
        return UrlMeta(method, path, param, potocol, metaStr)


@dataclass(frozen=True)
class LogLine:
    ipAddress: str
    identd: str
    userId: str
    receivedAt: dt.datetime
    url: str
    status_code: int
    return_size: int
    referer: str
    agent: str
    raw: str

    def dict(self):
        return {field.name: str(getattr(self, field.name)) for field in fields(self)}

    @staticmethod
    def read_log_lines(line: str):
        try:
            result = []
            result.append(line)

            I_IP = 0
            I_IID = 1
            I_UID = 2
            I_DATE = 3
            I_URL_S = 4
            I_URL_E = 5
            I_STATUS_S = 6
            I_STATUS_E = 7
            I_LENGTH = 8
            I_REFERER_S = 9
            I_REFERER_E = 10
            I_AGENT_S = 11
            I_AGENT_E = 12

            varIndex = 0
            buf = ""
            for c in line:
                if varIndex in [I_IP, I_IID, I_UID, I_STATUS_E, I_LENGTH] and c == " ":
                    if varIndex in [I_STATUS_E, I_LENGTH]:
                        result.append(int(buf))
                    else:
                        result.append(str(buf))
                    buf = ""
                    varIndex += 1
                elif varIndex in [I_DATE] and c == "[":
                    buf = ""
                elif varIndex in [I_DATE] and c == "]":
                    result.append(dt.datetime.strptime(buf, "%d/%b/%Y:%H:%M:%S %z"))
                    buf = ""
                    varIndex += 1
                elif varIndex in [I_URL_S, I_REFERER_S, I_AGENT_S] and c == "\"":
                    buf = ""
                    varIndex += 1
                elif varIndex in [I_URL_E] and c == "\"":
                    if buf[-1] == "\\":  # escaping
                        buf += c
                        continue

                    result.append(str(buf))
                    buf = ""
                    varIndex += 1
                elif varIndex in [I_REFERER_E, I_AGENT_E] and c == "\"":
                    result.append(str(buf))
                    buf = ""
                    varIndex += 1
                elif varIndex in [I_STATUS_S] and c == " ":
                    varIndex += 1
                else:
                    buf += c
            return result
        except Exception:
            print(line)
            raise

    @staticmethod
    def from_line(result: [any]) -> 'LogLine':
        ipaddress = result[LogLineIndex.IpAddress.value],
        identd = result[LogLineIndex.Identd.value],
        userId = result[LogLineIndex.UserId.value],
        receivedAt = result[LogLineIndex.ReceivedAt.value],
        url = result[LogLineIndex.Url.value],
        statusCode = result[LogLineIndex.StatusCode.value],
        returnSize = result[LogLineIndex.ReturnSize.value],
        referer = result[LogLineIndex.Referer.value],
        agent = result[LogLineIndex.Agent.value],
        raw = result[LogLineIndex.Raw.value]
        return LogLine(
            ipaddress, identd, userId, receivedAt, url, statusCode, returnSize, referer, agent, raw
        )

    @staticmethod
    def read_log_lines_and_parse(line) -> 'LogLine':
        return LogLine.from_line(LogLine.read_log_lines(line))





# -*- coding: utf-8 -*-
"""
Created on Wed Dec 21 08:34:20 2022

@author: adamh
"""
import datetime as dt
import json
import os
import sys
from dataclasses import dataclass
from enum import Enum
from typing import Callable, Generator, Tuple


class RenameHandler:

    def __init__(self, date: dt.datetime = dt.date.today(), separator: str = "-", strFormat: str = "%Y%m%d"):
        self.separator = separator
        self.strFormat = strFormat
        self.date = date

    def getFunction(self) -> Callable[[str], str]:
        def rename(file_name: str) -> str:
            return f"{file_name}{self.separator}{self.date.strftime(self.strFormat)}"

        return rename


class DeltaType(Enum):
    DATA = 'DATA'
    RENAMED = 'RENAMED'


@dataclass
class DeltaRead:
    deltaType: DeltaType
    data: str


class FileProgress:
    def __init__(self, line):
        lines = line.split("|")
        self.filename = lines[0]
        self.offset = int(lines[1])

    def setOffset(self, offset: int):
        self.offset = offset
        return self

    def __str__(self):
        return "%s|%d" % (self.filename, self.offset)

    def __repr__(self):
        return "[%s]" % str(self)

    def toJson(self):
        return str(self)


class ChangeHandler:
    @staticmethod
    def do_nothing(data: DeltaRead) -> DeltaRead:
        return data

    @staticmethod
    def print_changes(data: DeltaRead) -> DeltaRead:
        if data.deltaType == DeltaType.DATA and data.data != "":
            print(data.data, end="")
        elif data.deltaType == DeltaType.RENAMED:
            pass
        return data


class FileProgressUtils:
    def __init__(self, configFile="read-progress.json"):
        configFile = os.path.abspath(configFile)
        print(f"load config with abs file[{configFile}]")
        self.configFile = configFile
        self.loadConfig()

    def loadConfig(self):
        if not os.path.exists(self.configFile):
            with open(self.configFile, "w") as f:
                f.write("{}")

        with open(self.configFile, "r") as f:
            return json.load(f)

    def save_progress(self, fp: FileProgress):
        d = self.loadConfig()
        d.update({fp.filename: fp.toJson()})
        new_config = json.dumps(d, indent=4)
        with open(self.configFile, "w") as f:
            f.write(new_config)

    def getFileProgress(self, file_name: str) -> (FileProgress, bool):
        d = self.loadConfig()

        if file_name in d:
            return FileProgress(d[file_name]), False
        else:
            progress = FileProgress(f"{file_name}|0")
            self.save_progress(progress)
            return progress, True

    def setFileProgress(self, filename, offset):
        self.save_progress(self.getFileProgress(os.path.abspath(filename))[0].setOffset(offset))

    def checkOnce(self, filename) -> Generator[DeltaRead, None, None]:
        for item in self.checkOnceAndDo(filename):
            yield item

    def checkOnceAndDo(self, filename, renameHandler: Callable[[str], str] = None, min_read=1024*800) -> Generator[DeltaRead, None, None]:
        real_file = os.path.abspath(filename)
        (file_progress, createdNew) = self.getFileProgress(real_file)
        file_size = os.stat(filename).st_size
        if file_size < file_progress.offset:
            if renameHandler is None:
                _ = yield DeltaRead(DeltaType.RENAMED, None)
            else:
                gen = FileProgressUtils.read_all(renameHandler(file_progress.filename), file_progress.offset, min_read=min_read)
                while True:
                    try:
                        (rename_dr, _) = next(gen)
                        if rename_dr.deltaType == DeltaType.DATA:
                            yield rename_dr
                    except StopIteration:
                        break
                file_progress.setOffset(0)
                self.save_progress(file_progress)
        else:
            gen = FileProgressUtils.read_all(file_progress.filename, file_progress.offset, min_read=min_read)
            while True:
                try:
                    item = next(gen)
                    (read_dr, new_offset) = item
                    yield read_dr

                    file_progress.setOffset(new_offset)
                    self.save_progress(file_progress)
                except StopIteration:
                    break

    @staticmethod
    def read_all(filename: str, offset: int, min_read=1024 * 5) -> Generator[Tuple[DeltaRead, int], None, None]:
        real_file = os.path.abspath(filename)
        cur = 0
        lines = ""

        with open(real_file, "r") as f:
            f.seek(offset)
            # data = f.read()
            while True:
                data = f.readline()
                if len(data) == 0:
                    break
                if cur + len(data) >= min_read:
                    lines += data
                    yield DeltaRead(DeltaType.DATA, lines), f.tell()
                    lines = ""
                    cur = 0
                else:
                    cur += len(data)
                    lines += data

            if len(lines) > 0:
                yield DeltaRead(DeltaType.DATA, lines), f.tell()


if __name__ == "__main__":
    args = sys.argv
    fileName = args[1]
    if not os.path.exists(fileName):
        print("Commands: ")
        for arg in args:
            print(f"{arg}")
        raise Exception(f"File[{fileName}] not exists")
    fileName = os.path.abspath(fileName)
    dirName = os.path.dirname(fileName)
    print(f"File: {fileName} ")
    print(f"Dir: {dirName} ")

    fpu = FileProgressUtils()
    gen = fpu.checkOnceAndDo(fileName, RenameHandler(date=dt.date(2023, 1, 8), separator="_").getFunction())
    while True:
        try:
            dr = next(gen)
            if dr.deltaType == DeltaType.RENAMED:
                pass
            else:
                ChangeHandler.print_changes(dr)
        except StopIteration:
            pass



import re


class Ip:
    def __init__(self, ip_seg: [int]):
        self.valid(ip_seg)
        self.ip_seg = ip_seg

    def binary_notation(self):
        Ip.valid(self.ip_seg)
        ipStr = "".join(["{:08b}".format(seg) for seg in self.ip_seg[:-1]])
        if self.ip_seg[4] > -1:
            s = ""
            for k, i in enumerate(ipStr):
                if k > self.ip_seg[4]:
                    s += "X"
                else:
                    s += f"{i}"
            return s
        else:
            return ipStr

    @staticmethod
    def from_regular_form(ip_str: str) -> 'Ip':
        pattern = "(\d+)\.(\d+)\.(\d+)\.(\d+)(/\d+){0,1}"
        m = re.match(pattern, ip_str)
        return Ip([int(m[1]), int(m[2]), int(m[3]), int(m[4]), int(m[5][1:]) if m[5] else -1])

    @staticmethod
    def valid(ip_seg: [int]):
        for key, seg in enumerate(ip_seg):
            if key < 4:
                if seg > 255 or seg < 0:
                    raise f"IP{ip_seg} not valid for value: {seg}"
            else:
                if seg > 32 or seg < -1:
                    raise f"IP[{ip_seg}] CIDR not valid for value: {seg}"

    @staticmethod
    def regular_form(ip: [int]):
        s = ""
        return ".".join([f"{seg}" for key, seg in ip[:-1]]) + f"/{ip[-1]}" if ip[-1] > -1 else ""


if __name__ == "__main__":
    ip = Ip.from_regular_form("192.168.8.1/16")
    print(ip.binary_notation())
    ipResults = [
        print(f"{ipStr}[{pow(2, 32 - ip.ip_seg[4])}] {ip.binary_notation()}")
        for ipStr in
        "10.91.132.0/22\n10.91.136.0/21\n10.91.144.0/20\n10.91.160.0/19\n10.91.196.0/22\n10.91.200.0/21\n10.91.208.0/20\n10.91.224.0/19".split("\n")
        for ip in [Ip.from_regular_form(ipStr)]
    ]



import datetime as dt
from typing import Callable


class ProgressPinger:
    def __init__(self, print_every_n_count: int = 100, print_every_n_second: int = 15):
        self.count = 0
        self.lastUpdate = dt.datetime.now()
        self.print_every_n_count = print_every_n_count
        self.print_every_n_second = print_every_n_second

    def ping(self):
        self.count += 1

        printed = False
        if self.count % self.print_every_n_count == 0:
            ProgressPinger.print_log(f"Processed count: {self.print_every_n_count}")
            printed = True

        if not printed and self.lastUpdate + dt.timedelta(seconds=self.print_every_n_second) < dt.datetime.now():
            ProgressPinger.print_log(f"Processed count: {self.print_every_n_count}")

        self.lastUpdate = dt.datetime.now()

    @staticmethod
    def log_formatting(msg: str) -> str:
        return f"[{dt.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] {msg}"

    @staticmethod
    def print_log(msg: str, formatting: Callable[[str], str] = log_formatting):
        print(formatting(msg))



from socket import socket


class IpHostFinder:
    map = {}

    def find(self, ip_str: str):
        if ip_str in self.map:
            return self.map[ip_str]
        else:
            try:
                domain, s = socket.getnameinfo((ip_str, 0), 0)
                self.map.update({ip_str: domain})
                return domain
            except:
                self.map.update({ip_str: "Unknown"})
                return "Unknown"
